from http import HTTPStatus

from pytest_sanic.utils import TestClient
from typing import Dict, Text, List

import pytest
from sanic.testing import SanicTestClient
from sqlalchemy.orm import Session

from rasax.community.api.blueprints.stack import TAGS_ANY_REQUEST_PARAMETER
from rasax.community.database.conversation import Conversation
from tests.unit.api.blueprints.conftest import (
    get_conversation_by_id,
    CONVERSATIONS_ENDPOINT,
    TAGS_ENDPOINT,
    tag_1,
    tag_2,
    tag_3,
)


async def test_events_metadata_id(
    test_client: SanicTestClient,
    auth_header: Dict[Text, Text],
    session: Session,
    conversation: Text,
):
    response = await test_client.get(
        f"{CONVERSATIONS_ENDPOINT}/{conversation}", headers=auth_header
    )
    response_body = await response.json()
    received_ids = [e["metadata"]["rasa_x_id"] for e in response_body["events"]]

    conversation_entity = session.query(Conversation).get(conversation)
    expected_ids = [e.id for e in conversation_entity.events]

    assert received_ids == expected_ids


async def test_events_metadata_is_flagged(
    test_client: SanicTestClient,
    auth_header: Dict[Text, Text],
    session: Session,
    conversation: Text,
):
    response = await test_client.get(
        f"{CONVERSATIONS_ENDPOINT}/{conversation}", headers=auth_header
    )
    response_body = await response.json()
    received_flagged_ids = [
        e["metadata"]["rasa_x_id"]
        for e in response_body["events"]
        if e["metadata"]["rasa_x_flagged"]
    ]

    conversation_entity = session.query(Conversation).get(conversation)
    expected_flagged_ids = [e.id for e in conversation_entity.events if e.is_flagged]

    assert received_flagged_ids == expected_flagged_ids


@pytest.mark.parametrize(
    "exclude_leading_action_session_start,expected_timestamps",
    [("true", [2000, 3000, 4000, 5000]), ("false", [1000, 2000, 3000, 4000, 5000])],
)
async def test_events_with_exclude_action_session_start(
    test_client: SanicTestClient,
    auth_header: Dict[Text, Text],
    conversation: Text,
    exclude_leading_action_session_start: Text,
    expected_timestamps: List[float],
):
    response = await test_client.get(
        f"{CONVERSATIONS_ENDPOINT}/{conversation}",
        headers=auth_header,
        params={
            "exclude_leading_action_session_start": exclude_leading_action_session_start
        },
    )
    response_body = await response.json()

    assert [
        event["timestamp"] for event in response_body["events"]
    ] == expected_timestamps


@pytest.mark.parametrize(
    "accept_header_value,output_contains,content_type",
    [
        ("", 'sender_id":"{conversation_id}"', "application/json"),
        ("application/json", 'sender_id":"{conversation_id}"', "application/json"),
        ("text/markdown", "* test\n    - action\n", "text/markdown"),
    ],
)
async def test_get_events_with_dumps(
    test_client: TestClient,
    auth_header: Dict[Text, Text],
    conversation: Text,
    accept_header_value: Text,
    output_contains: Text,
    content_type: Text,
):
    headers = auth_header.copy()
    headers.update({"Accept": accept_header_value})
    response = await test_client.get(
        f"{CONVERSATIONS_ENDPOINT}/{conversation}", headers=headers,
    )

    # check status code
    assert response.status == 200

    # make sure content_type header is correct
    assert response.headers.get("Content-Type") == content_type

    # check content of request body
    expected_content = output_contains.format(conversation_id=conversation)
    output = await response.text()
    assert expected_content in output


async def test_get_tags_for_conversation(
    conversation: Text, test_client: TestClient, auth_header: Dict[Text, Text],
):
    post_url = f"{CONVERSATIONS_ENDPOINT}/{conversation}/{TAGS_ENDPOINT}"
    get_url = f"{CONVERSATIONS_ENDPOINT}"

    response = await test_client.get(get_url, headers=auth_header)
    response_body = await response.json()
    assert len(get_conversation_by_id(response_body, conversation)["tags"]) == 0

    await test_client.post(post_url, headers=auth_header, json=tag_1)

    response = await test_client.get(get_url, headers=auth_header)
    response_body = await response.json()
    assert len(get_conversation_by_id(response_body, conversation)["tags"]) == 1

    await test_client.post(post_url, headers=auth_header, json=tag_2)

    response = await test_client.get(get_url, headers=auth_header)
    response_body = await response.json()
    assert len(get_conversation_by_id(response_body, conversation)["tags"]) == 2


async def test_get_tags_for_conversation_not_admin_role(
    conversation_for_tester: Text,
    test_client: TestClient,
    auth_header_tester: Dict[Text, Text],
):
    post_url = f"{CONVERSATIONS_ENDPOINT}/{conversation_for_tester}/{TAGS_ENDPOINT}"
    get_url = f"{CONVERSATIONS_ENDPOINT}"

    response = await test_client.get(get_url, headers=auth_header_tester)
    response_body = await response.json()
    assert (
        len(get_conversation_by_id(response_body, conversation_for_tester)["tags"]) == 0
    )

    await test_client.post(post_url, headers=auth_header_tester, json=tag_1)

    response = await test_client.get(get_url, headers=auth_header_tester)
    response_body = await response.json()
    assert (
        len(get_conversation_by_id(response_body, conversation_for_tester)["tags"]) == 1
    )


async def test_get_tags_for_multiple_conversations(
    conversation: Text,
    conversation_other: Text,
    test_client: TestClient,
    auth_header: Dict[Text, Text],
):
    tags_endpoint_for_conversation1 = (
        f"{CONVERSATIONS_ENDPOINT}/{conversation}/{TAGS_ENDPOINT}"
    )
    tags_endpoint_for_conversation2 = (
        f"{CONVERSATIONS_ENDPOINT}/{conversation_other}/{TAGS_ENDPOINT}"
    )
    get_url = f"{CONVERSATIONS_ENDPOINT}"

    response = await test_client.get(get_url, headers=auth_header)
    response_body = await response.json()
    assert len(get_conversation_by_id(response_body, conversation)["tags"]) == 0
    assert len(get_conversation_by_id(response_body, conversation_other)["tags"]) == 0

    await test_client.post(
        tags_endpoint_for_conversation1, headers=auth_header, json=tag_1
    )

    await test_client.post(
        tags_endpoint_for_conversation2, headers=auth_header, json=tag_2
    )

    response = await test_client.get(get_url, headers=auth_header)
    response_body = await response.json()
    assert len(get_conversation_by_id(response_body, conversation)["tags"]) == 1
    assert len(get_conversation_by_id(response_body, conversation_other)["tags"]) == 1


async def test_filter_tag_assigned_to_both_conversations(
    conversation: Text,
    conversation_other: Text,
    test_client: TestClient,
    auth_header: Dict[Text, Text],
):
    tags_endpoint_for_conversation1 = (
        f"{CONVERSATIONS_ENDPOINT}/{conversation}/{TAGS_ENDPOINT}"
    )
    tags_endpoint_for_conversation2 = (
        f"{CONVERSATIONS_ENDPOINT}/{conversation_other}/{TAGS_ENDPOINT}"
    )
    get_url_template = f"{CONVERSATIONS_ENDPOINT}?{TAGS_ANY_REQUEST_PARAMETER}="

    response = await test_client.post(
        tags_endpoint_for_conversation1, headers=auth_header, json=tag_1
    )
    response_body = await response.json()
    tag_1_id = response_body[0]["id"]

    await test_client.post(
        tags_endpoint_for_conversation2, headers=auth_header, json=tag_1
    )

    url_filter_1 = f"{get_url_template}{tag_1_id}"
    response = await test_client.get(url_filter_1, headers=auth_header)
    response_body = await response.json()
    assert get_conversation_by_id(response_body, conversation)
    assert get_conversation_by_id(response_body, conversation_other)


async def test_filter_tag_assigned_to_only_one_conversation(
    conversation: Text,
    conversation_other: Text,
    test_client: TestClient,
    auth_header: Dict[Text, Text],
):
    tags_endpoint_for_conversation1 = (
        f"{CONVERSATIONS_ENDPOINT}/{conversation}/{TAGS_ENDPOINT}"
    )
    tags_endpoint_for_conversation2 = (
        f"{CONVERSATIONS_ENDPOINT}/{conversation_other}/{TAGS_ENDPOINT}"
    )
    get_url_template = f"{CONVERSATIONS_ENDPOINT}?{TAGS_ANY_REQUEST_PARAMETER}="

    await test_client.post(
        tags_endpoint_for_conversation1, headers=auth_header, json=tag_1
    )

    response = await test_client.post(
        tags_endpoint_for_conversation1, headers=auth_header, json=tag_2
    )
    response_body = await response.json()
    tag_2_id = response_body[0]["id"]

    await test_client.post(
        tags_endpoint_for_conversation2, headers=auth_header, json=tag_1
    )

    url_filter_2 = f"{get_url_template}{tag_2_id}"
    response = await test_client.get(url_filter_2, headers=auth_header)
    response_body = await response.json()
    assert get_conversation_by_id(response_body, conversation)
    assert not get_conversation_by_id(response_body, conversation_other)


async def test_filter_any_tag_is_assigned_to_conversation(
    conversation: Text,
    conversation_other: Text,
    test_client: TestClient,
    auth_header: Dict[Text, Text],
):
    tags_endpoint_for_conversation1 = (
        f"{CONVERSATIONS_ENDPOINT}/{conversation}/{TAGS_ENDPOINT}"
    )
    tags_endpoint_for_conversation2 = (
        f"{CONVERSATIONS_ENDPOINT}/{conversation_other}/{TAGS_ENDPOINT}"
    )
    get_url_template = f"{CONVERSATIONS_ENDPOINT}?{TAGS_ANY_REQUEST_PARAMETER}="

    await test_client.post(
        tags_endpoint_for_conversation1, headers=auth_header, json=tag_1
    )

    response = await test_client.post(
        tags_endpoint_for_conversation1, headers=auth_header, json=tag_2
    )
    response_body = await response.json()
    tag_2_id = response_body[0]["id"]

    await test_client.post(
        tags_endpoint_for_conversation2, headers=auth_header, json=tag_1
    )

    response = await test_client.post(
        tags_endpoint_for_conversation2, headers=auth_header, json=tag_3
    )
    response_body = await response.json()
    tag_3_id = response_body[0]["id"]

    url_filter_4 = f"{get_url_template}{tag_2_id},{tag_3_id}"
    response = await test_client.get(url_filter_4, headers=auth_header)
    response_body = await response.json()
    assert get_conversation_by_id(response_body, conversation)
    assert get_conversation_by_id(response_body, conversation_other)


async def test_filter_tag_not_assigned_to_any_conversation(
    conversation: Text,
    conversation_other: Text,
    test_client: TestClient,
    auth_header: Dict[Text, Text],
):
    tags_endpoint_for_conversation1 = (
        f"{CONVERSATIONS_ENDPOINT}/{conversation}/{TAGS_ENDPOINT}"
    )
    tags_endpoint_for_conversation2 = (
        f"{CONVERSATIONS_ENDPOINT}/{conversation_other}/{TAGS_ENDPOINT}"
    )
    get_url_template = f"{CONVERSATIONS_ENDPOINT}?{TAGS_ANY_REQUEST_PARAMETER}="

    await test_client.post(
        tags_endpoint_for_conversation1, headers=auth_header, json=tag_1
    )

    await test_client.post(
        tags_endpoint_for_conversation1, headers=auth_header, json=tag_2
    )

    await test_client.post(
        tags_endpoint_for_conversation2, headers=auth_header, json=tag_1
    )

    await test_client.post(
        tags_endpoint_for_conversation2, headers=auth_header, json=tag_3
    )

    url_filter_5 = f"{get_url_template}999"
    response = await test_client.get(url_filter_5, headers=auth_header)
    response_body = await response.json()
    assert not get_conversation_by_id(response_body, conversation)
    assert not get_conversation_by_id(response_body, conversation_other)
