import os
import webbrowser
from pathlib import Path
from typing import Dict, Text
from unittest.mock import Mock

import pytest
from _pytest.monkeypatch import MonkeyPatch
from sqlalchemy.orm import Session

from rasa.utils import common as rasa_utils
from rasax.community import initialise, config, utils
from rasax.community.constants import (
    CONFIG_TELEMETRY_ENABLED,
    CONFIG_TELEMETRY_WELCOME_SHOWN,
    WELCOME_PAGE_URL,
)
from rasax.community.services.data_service import DataService
from rasax.community.services.settings_service import SettingsService
from rasax.community.services.story_service import StoryService
from rasax.community.services.user_service import UserService
from tests.unit.conftest import insert_test_user

EXAMPLE_URL = "https://example.com"


def test_create_community_user(
    settings_service: SettingsService, session: Session, user_service: UserService
):
    from rasax.community.constants import COMMUNITY_USERNAME

    # Change this to `False` for the test to make sure our test password is used
    config.LOCAL_MODE = False

    test_password = "123456"
    settings_service.save_community_user_password(test_password)

    initialise.create_community_user(session, Mock())

    actual = user_service._fetch_user(COMMUNITY_USERNAME)

    assert actual is not None
    assert actual.password_hash == user_service.hash_pw(test_password)

    # Change it back
    config.LOCAL_MODE = True
    user_service.delete_user(COMMUNITY_USERNAME)


def test_create_community_user_with_password_from_env(
    session: Session, user_service: UserService
):
    from rasax.community.constants import COMMUNITY_USERNAME, ENV_RASA_X_PASSWORD

    password = "123456"
    os.environ[ENV_RASA_X_PASSWORD] = password

    # Change this to `False` for the test to make sure our test password is used
    config.LOCAL_MODE = False

    # create user
    initialise.create_community_user(session, Mock())
    created_user = user_service._fetch_user(COMMUNITY_USERNAME)
    assert created_user is not None
    assert created_user.password_hash == user_service.hash_pw(password)

    # Change it back
    config.LOCAL_MODE = True
    user_service.delete_user(COMMUNITY_USERNAME)


def test_attempt_to_recreate_community_user(
    session: Session, user_service: UserService
):
    from rasax.community.constants import COMMUNITY_USERNAME

    # Change this to `False` for the test to make sure our test password is used
    config.LOCAL_MODE = False

    # create user
    initialise.create_community_user(session, Mock())
    created_user = user_service._fetch_user(COMMUNITY_USERNAME)
    assert created_user is not None
    password1 = created_user.password_hash

    # try to create it second time
    initialise.create_community_user(session, Mock())
    created_user = user_service._fetch_user(COMMUNITY_USERNAME)
    assert created_user is not None
    password2 = created_user.password_hash

    # it means password has not been changed
    assert password1 == password2

    # Change it back
    config.LOCAL_MODE = True
    user_service.delete_user(COMMUNITY_USERNAME)


def test_attempt_to_recreate_community_user_with_password_from_env(
    session: Session, user_service: UserService
):
    from rasax.community.constants import COMMUNITY_USERNAME, ENV_RASA_X_PASSWORD

    password = "123456"
    os.environ[ENV_RASA_X_PASSWORD] = password

    # Change this to `False` for the test to make sure our test password is used
    config.LOCAL_MODE = False

    # create user
    initialise.create_community_user(session, Mock())
    created_user = user_service._fetch_user(COMMUNITY_USERNAME)
    assert created_user is not None
    password1 = created_user.password_hash
    assert password1 == user_service.hash_pw(password)

    # try to create it second time
    password = "654321"
    os.environ[ENV_RASA_X_PASSWORD] = password

    initialise.create_community_user(session, Mock())
    created_user = user_service._fetch_user(COMMUNITY_USERNAME)
    assert created_user is not None
    password2 = created_user.password_hash

    # it means password has not been changed
    assert password1 == password2

    # Change it back
    config.LOCAL_MODE = True
    user_service.delete_user(COMMUNITY_USERNAME)


@pytest.mark.parametrize(
    "telemetry_config, expected_url",
    [
        (
            # Telemetry enabled, welcome page not shown yet
            {CONFIG_TELEMETRY_ENABLED: True, CONFIG_TELEMETRY_WELCOME_SHOWN: False},
            WELCOME_PAGE_URL,
        ),
        (
            # Telemetry enabled, but upgrading from an older Rasa X version
            # Don't show welcome page
            {CONFIG_TELEMETRY_ENABLED: True},
            EXAMPLE_URL,
        ),
        (
            # Telemetry enabled, welcome page already shown
            {CONFIG_TELEMETRY_ENABLED: True, CONFIG_TELEMETRY_WELCOME_SHOWN: True},
            EXAMPLE_URL,
        ),
        (
            # Telemetry not enabled
            {CONFIG_TELEMETRY_ENABLED: False},
            EXAMPLE_URL,
        ),
    ],
)
def test_open_web_browser(
    monkeypatch: MonkeyPatch, telemetry_config: Dict[Text, bool], expected_url: Text
):
    open_new_tab = Mock()
    monkeypatch.setattr(webbrowser, "open_new_tab", open_new_tab)

    read_global_config_value = Mock()
    read_global_config_value.return_value = telemetry_config
    monkeypatch.setattr(
        rasa_utils, "read_global_config_value", read_global_config_value
    )

    write_global_config_value = Mock()
    monkeypatch.setattr(
        rasa_utils, "write_global_config_value", write_global_config_value
    )

    initialise.AppStartedCallable.open_web_browser(EXAMPLE_URL)

    # Check which URL was opened first
    assert open_new_tab.call_args_list[0][0][0] == expected_url


async def test_inject_files_from_disk_does_not_dump(
    tmpdir: Path, new_project: Text, session: Session, monkeypatch: MonkeyPatch
):
    from rasa.cli import scaffold

    username = "user 919"
    insert_test_user(session, username)

    # Create Rasa project
    project_path = str(tmpdir)
    scaffold.create_initial_project(project_path)

    dump = Mock()

    monkeypatch.setattr(utils, "should_dump", lambda: True)
    monkeypatch.setattr(SettingsService, "_dump_config", dump)
    monkeypatch.setattr(DataService, "_dump_data", dump)
    monkeypatch.setattr(StoryService, "_dump_stories", dump)
    await initialise.inject_files_from_disk(
        project_path, "data", session, username=username
    )

    dump.assert_not_called()
